/* SetManList 2.1 */

OPT PREPROCESS
OPT OSVERSION=37

MODULE 'intuition/gadgetclass','intuition/icclass','libraries/gadtools',
       'graphics/gfxbase','graphics/text',
       'intuition','graphics','dos','intuition/classes','intuition/classusr',
       'utility','gadtools','intuition/screens','exec/ports','exec/semaphores',
       'exec/lists','exec/nodes','dos/dos','dos/dosextens','exec/libraries',
       'exec/tasks','dos/dosasl','amigalib/mempools','exec/memory'


OBJECT setmanport
   mp:mp
   version[4]:ARRAY
   sem:ss
   lh:mlh
ENDOBJECT


OBJECT setnode
   n:mln
   lib:PTR TO lib
   offset
   task:PTR TO tc
   taskname[22]:ARRAY
   jmp1:INT
   setfunc
   jmp2:INT
   prevfunc
   nextfunc
ENDOBJECT


OBJECT libraryinfo
   next:PTR TO libraryinfo
   libraryname[30]:ARRAY
   entries:INT
   functionname:PTR TO CHAR  /*First ptr */
ENDOBJECT



#define VERSION '2.0'	/* must be 3 characters */

#define SFOFFSET		(-$01A4)	/* SetFunction() vector */

CONST JMPCODE =	$4ef9
CONST BRACODE =	$6004
CONST NOPCODE =	$4e71


DEF smport:PTR TO setmanport
DEF libinfolist=NIL:PTR TO libraryinfo
DEF pool=NIL


PROC main() HANDLE
DEF argarray:PTR TO LONG, rdargs=NIL

   Forbid()
   IF (smport := FindPort('SetMan')) = NIL
      Permit()
      WriteF('SetMan is not running.\n')
      RETURN 20
   ENDIF

   Permit()

   IF (utilitybase := OpenLibrary('utility.library',37)) = NIL THEN Raise("LIB")

   IF Stricmp(smport.version,VERSION)<> 0
      WriteF('This version is not compatible with the running SetMan version!\n')
      Raise("VERS")
   ENDIF

   IF (pool := libCreatePool(MEMF_ANY OR MEMF_CLEAR,10000,500))=NIL THEN Raise("MEM")

      argarray := [0,'FD:',0]
      IF (rdargs := ReadArgs('NOFD/S, FDPATH',argarray,NIL)) = NIL
         PrintFault(IoErr(),NIL)
         Raise("ARGS")
      ENDIF

   WriteF('SetManList 2.1 1997-98 by Eric Sauvageau.\n\n')

   IF argarray[0] =NIL THEN readAllFD(argarray[1])

   showlist()

EXCEPT DO

   IF rdargs THEN FreeArgs(rdargs)
   IF pool THEN libDeletePool(pool)

   CloseLibrary(utilitybase)

ENDPROC




PROC showlist()
DEF node:PTR TO setnode
DEF libinf:PTR TO libraryinfo
DEF count=0, abort=FALSE

DEF st1[50]:STRING,
    st2[50]:STRING,
    st3:PTR TO CHAR,
    offs,
    statname:PTR TO LONG,
    functn:PTR TO CHAR


   WriteF('Handler Task           Code       Function                          State\n\s',{asciiline})

   ObtainSemaphore(smport.sem)

   node := smport.lh.head


   statname := ['Open()','Close()','Expunge()','ExtFunc()','BeginIO()','AbortIO()']

   WHILE (node.n.succ <> NIL) AND ((abort := CtrlC())=FALSE)

      INC count

      offs := -node.offset/6
      
      IF ((offs <= 4) AND (node.lib.ln.type = NT_LIBRARY)) OR ((offs <= 6) AND (node.lib.ln.type = NT_DEVICE))
         StringF(st2,' \s \s ',statname[offs-1],node.lib.ln.name)

      ELSEIF (libinf := findLibraryInfo(libinfolist, node.lib.ln.name))
         functn := libinf + (SIZEOF libraryinfo)

         IF (offs <= libinf.entries) AND (Long(functn + (offs*4)))
/* Using \c for the ( because of StringF() trying to interpret it */
            StringF(st2,' \s\c)',(Long(functn + (offs*4))),"(")
         ENDIF
      ELSE
         StringF(st2,' -$\z\h[4] \s ',(0-node.offset) AND $FFFF,node.lib.ln.name)
      ENDIF

      IF (node.setfunc = NIL)
         st3 := 'Removed'
      ELSEIF (node.jmp1 = JMPCODE)
         st3 := 'Active'
      ELSE
         st3 := 'Disabled'
      ENDIF

      WriteF('\l\s[22]\s[10] \s[35]\s[8]\n',
                                   node.taskname, 
                                   (IF node.setfunc THEN StringF(st1,' $\z\h[8] ',node.setfunc) ELSE '  -------- '),
                                   st2,
                                   st3)

      node := node.n.succ
   ENDWHILE

   ReleaseSemaphore(smport.sem)

   IF abort
      PrintFault(ERROR_BREAK, NIL)
   ELSE
      WriteF('\n\d patches installed.\n',count)
   ENDIF
ENDPROC


/*** FD file parsing ***/


PROC readAllFD(fdpath:PTR TO CHAR) HANDLE
DEF next:PTR TO libraryinfo,
    fib=NIL:PTR TO fileinfoblock,
    dir=NIL, odir=NIL,
    proc:PTR TO process,
    winptr,
    lib=NIL:PTR TO libraryinfo,
    count=0, ptrascii:PTR TO LONG


proc := FindTask(NIL)
winptr := proc.windowptr

proc.windowptr := -1

dir := Lock(fdpath,ACCESS_READ)

proc.windowptr := winptr

ptrascii := ['\\','|','/','-']

IF (dir)
   WriteF('Parsing FD files... \e[0 p')
   odir := CurrentDir(dir)

   IF (fib := AllocDosObject(DOS_FIB,NIL))
      IF (Examine(dir,fib))
         next := libinfolist

         WHILE (ExNext(dir,fib))
            IF (next := readFD(fib.filename))

               WriteF('\s\e[D',ptrascii[count++])
               IF count = 4 THEN count := 0
               IF lib = NIL
                  lib := next
                  libinfolist := lib
               ELSE 
                  lib.next := next
                  lib := lib.next
               ENDIF
            ENDIF
         ENDWHILE
      ENDIF
   ENDIF

ENDIF

EXCEPT DO
   IF fib THEN FreeDosObject(DOS_FIB,fib)
   IF odir THEN CurrentDir(odir)
   IF dir THEN UnLock(dir)

   WriteF('\b\e[1 p')

ENDPROC




PROC readFD(name:PTR TO CHAR)
DEF bptr,
    libname[35]:STRING,
    tmpstr[35]:STRING,
    lib=NIL:PTR TO libraryinfo


/* only names ending in '_lib.fd' allowed */
IF InStr(name,'_lib.fd',0)=-1 THEN RETURN NIL


   IF (bptr := Open(name,MODE_OLDFILE))

      IF (lib := processFD(bptr))

         /* Shove real name in lib */
         IF Strnicmp(name,'wb',STRLEN)=0
            StrCopy(libname,'workbench.library')
         ELSE
            StringF(libname,'\s.library',StrCopy(tmpstr,name,InStr(name,'_lib.fd')))
         ENDIF
         AstrCopy(lib.libraryname,libname,28)
      ENDIF

	   Close(bptr)
   ENDIF
ENDPROC lib




PROC processFD(handle) HANDLE
DEF entries = 0, buffer[200]:ARRAY
DEF lib=NIL:PTR TO libraryinfo
DEF offset:PTR TO LONG

DEF ptr:PTR TO CHAR

/* Count how many lines we have */
   WHILE (Fgets(handle,buffer,60) <> NIL)
      IF Strnicmp(buffer,'##bias',6)=0 THEN entries := entries + (Val(buffer+6)/6) ELSE entries++
   ENDWHILE

/* Restore pointer to start of file. */
   Seek(handle,0, OFFSET_BEGINNING)

/* +4 for safety reasons, until I can finetune it */
   IF (lib := libAllocPooled(pool,(SIZEOF libraryinfo) + (entries+4)*4))=NIL THEN Raise("MEM")

/* Re-reading it.  Since our file is small, we're relying on the 
   filesystem's buffering to avoid unnecessary disk access rather than 
   reading and parsing twice from our own buffers.  Not the fastest, but 
   surely the simplest. */

   offset := lib + (SIZEOF libraryinfo)

   WHILE (Fgets(handle,buffer,200) <> NIL)
/* Hardcoded to 60 chars - will have to make it dynamic */

      IF (buffer[0] <> "*") AND (buffer[0] <> "\n")
         IF buffer[0] = "#"
            IF Strnicmp(buffer, '##bias',6)=0
               offset := lib + (SIZEOF libraryinfo) + ((Val(buffer+7)/6)*4)
            ENDIF
         ELSE
            IF (ptr := libAllocPooled(pool,60))=NIL THEN Raise("MEM")

            offset[] := ptr
            AstrCopy(ptr, buffer, InStr(buffer,'(')+1)
            offset++
         ENDIF
      ENDIF
   ENDWHILE

   lib.entries := (offset - lib - (SIZEOF libraryinfo)) / 4
EXCEPT

ENDPROC lib


PROC findLibraryInfo(list:PTR TO libraryinfo, name:PTR TO CHAR)

   WHILE (list)
      IF Stricmp(list.libraryname,name)=0 THEN RETURN list
      list := list.next
   ENDWHILE
ENDPROC NIL



CHAR '$VER: SetManList 2.1 (21.5.98)',0

asciiline:
CHAR '----------------------------------------------------------------------------\n',0
